from pathlib import Path
from textwrap import dedent

import aiofiles

__func_alias__ = {"list_": "list"}


async def list_(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    await hub.idemenv.ops.fill_local_version_list()
    current_version = await hub.idemenv.ops.get_current_version()
    version_list = []
    for ver in sorted(hub.idemenv.ops.LOCAL_VERSIONS.keys(), reverse=True):
        prefix = "  "
        suffix = ""
        if ver == current_version[0]:
            prefix = "* "
            suffix = f" (set by {current_version[1]})"
        version_list.append(prefix + ver + suffix)
    print("\n".join(version_list))


async def list_remote(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    await hub.idemenv.ops.fill_remote_version_list()
    version_list = sorted(hub.idemenv.ops.REMOTE_VERSIONS.keys(), reverse=True)
    if "latest" in version_list:
        version_list.pop(version_list.index("latest"))
    print("\n".join(version_list))


async def install(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    await hub.idemenv.ops.fill_remote_version_list()
    await hub.idemenv.ops.fill_local_version_list()

    if hub.OPT.idemenv.idem_version in hub.idemenv.ops.LOCAL_VERSIONS:
        print(f"{hub.OPT.idemenv.idem_version} is already installed.")
    elif hub.OPT.idemenv.idem_version in hub.idemenv.ops.REMOTE_VERSIONS:
        await hub.idemenv.ops.download_version(hub.OPT.idemenv.idem_version)
    else:
        print(f"ERROR: {hub.OPT.idemenv.idem_version} is not available as a binary.")


async def uninstall(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    await hub.idemenv.ops.fill_remote_version_list()
    await hub.idemenv.ops.fill_local_version_list()

    if hub.OPT.idemenv.idem_version in hub.idemenv.ops.LOCAL_VERSIONS:
        await hub.idemenv.ops.remove_version(hub.OPT.idemenv.idem_version)
    else:
        print(f"{hub.OPT.idemenv.idem_version} is already uninstalled.")


async def use(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    await hub.idemenv.ops.use_version(hub.OPT.idemenv.idem_version)


async def pin(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    await hub.idemenv.ops.pin_current_version()


async def version(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    current_version = await hub.idemenv.ops.get_current_version()
    if current_version[0]:
        print(f"{current_version[0]} (set by {current_version[1]})")
    else:
        print("ERROR: No version of Idem is set!")


async def init(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    idem_bin = Path(hub.OPT.idemenv.idemenv_dir) / "bin" / "idem"
    idem_bin.parent.mkdir(parents=True, exist_ok=True)

    wrapper = dedent(
        f"""
        #!/usr/bin/env python3

        import os
        import subprocess
        import sys
        from pathlib import Path

        if __name__ == "__main__":
            idemenv_dir = Path("{hub.OPT.idemenv.idemenv_dir}")
            override_version_file = Path(os.getcwd()) / ".idem-version"
            main_version_file = idemenv_dir / "version"

            current_version = ""

            if override_version_file.exists():
                with open(override_version_file, "r") as cfile:
                    current_version = cfile.read()
                    current_version = current_version.replace("\\n", "").strip()

            if not current_version and main_version_file.exists():
                with open(main_version_file, "r") as vfile:
                    current_version = vfile.read()
                    current_version = current_version.replace("\\n", "").strip()

            current_bin = idemenv_dir / "versions" / ("idem-" + current_version)

            if current_version and current_bin.exists():
                del sys.argv[0]
                cmd = (
                    [
                        str(current_bin),
                    ]
                    + sys.argv
                )
                subprocess.call(cmd)
            elif current_version:
                print("ERROR: Version " + current_version + " of Idem is not installed!")
            else:
                print("ERROR: No version of Idem specified!")
        """
    )
    wrapper = wrapper[1:]

    async with aiofiles.open(idem_bin, "w") as ofile:
        await ofile.write(wrapper)
    idem_bin.chmod(0o755)

    print(
        "Add the idemenv bin directory to your PATH:\n\n"
        f"    echo 'export PATH=\"{idem_bin.parent}:$PATH\"' >> ~/.bashrc\n"
        "OR:\n"
        f"    echo 'export PATH=\"{idem_bin.parent}:$PATH\"' >> ~/.zshrc\n"
    )



import os
import re
import sys
from pathlib import Path
from types import SimpleNamespace

import aiofiles
from bs4 import BeautifulSoup

LOCAL_VERSIONS = {}
REMOTE_VERSIONS = {}


async def local_version_list(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    ret = {}
    versions_dir = Path(hub.OPT.idemenv.idemenv_dir) / "versions"
    if versions_dir.exists():
        ret = versions_dir.glob("idem-*")
        ret = {ver.name.replace("idem-", ""): ver for ver in ret}
    return ret


async def fill_local_version_list(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    hub.idemenv.ops.LOCAL_VERSIONS = await hub.idemenv.ops.local_version_list()


async def remote_version_list(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    ret = {}
    ctx = SimpleNamespace(acct={})
    repo_list = await hub.exec.request.raw.get(
        ctx,
        url=hub.OPT.idemenv.repo_url,
    )
    if repo_list.get("ret"):
        soup = BeautifulSoup(repo_list["ret"], "html.parser")
        ret = {
            re.sub(r"-\d+$", "", node["href"][:-1]): node["href"][:-1]
            for node in soup.find_all("a")
            if node.get("href") and node["href"].endswith("/") and node["href"] != "../"
        }
    return ret


async def fill_remote_version_list(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    hub.idemenv.ops.REMOTE_VERSIONS = await hub.idemenv.ops.remote_version_list()


async def download_version(hub, version, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    ret = False
    ctx = SimpleNamespace(acct={})
    if not hub.idemenv.ops.REMOTE_VERSIONS:
        await hub.idemenv.ops.fill_remote_version_list()

    if version in hub.idemenv.ops.REMOTE_VERSIONS:
        url = hub.OPT.idemenv.repo_url
        if not url.endswith("/"):
            url += "/"
        url += hub.idemenv.ops.REMOTE_VERSIONS[version]
        file_list = await hub.exec.request.raw.get(
            ctx,
            url=url,
        )
        soup = BeautifulSoup(file_list["ret"], "html.parser")
        links = [
            node["href"]
            for node in soup.find_all("a")
            if node.get("href") and not node["href"].endswith("/") and node["href"] != "../"
        ]

        pkg_name = ""
        # TODO: verify download with SHA/GPG
        for link in links:
            if sys.platform in link:
                pkg_name = link

        if pkg_name:
            outfile = Path(hub.OPT.idemenv.idemenv_dir) / "downloads" / pkg_name
            outfile.parent.mkdir(parents=True, exist_ok=True)
            versions_dir = Path(hub.OPT.idemenv.idemenv_dir) / "versions"
            versions_dir.mkdir(parents=True, exist_ok=True)
            idem_bin_out = versions_dir / f"idem-{version}"

            download_url = "/".join(
                [
                    url,
                    pkg_name,
                ]
            )

            pkg = {}
            if not outfile.exists():
                pkg = await hub.exec.request.raw.get(
                    ctx,
                    url=download_url,
                )
                async with aiofiles.open(outfile, "wb") as ofile:
                    await ofile.write(pkg["ret"])

            if (outfile.exists() and not idem_bin_out.exists()) or (pkg and pkg["status"] == 200):
                print("Processing download...")
                outfile.rename(idem_bin_out)
                idem_bin_out.chmod(0o755)
                ret = idem_bin_out.exists()
    return ret


async def get_current_version(hub, **kwargs):
    """
    Read the current active version from ./.idem-version and the main version
    file at ${SALTENV_DIR}/version. The former overrides the latter.
    """
    ret = ("", "")
    override_version_file = Path(os.getcwd()) / ".idem-version"
    main_version_file = Path(hub.OPT.idemenv.idemenv_dir) / "version"

    if override_version_file.exists():
        async with aiofiles.open(override_version_file, "r") as cfile:
            current_version = await cfile.read()
            current_version = current_version.replace("\n", "").strip()
        ret = (current_version, str(override_version_file))

    if not ret[0] and main_version_file.exists():
        async with aiofiles.open(main_version_file, "r") as vfile:
            current_version = await vfile.read()
            current_version = current_version.replace("\n", "").strip()
        ret = (current_version, str(main_version_file))

    return ret


async def pin_current_version(hub, **kwargs):
    """
    Get the current active version and write it to ./.idem-version
    """
    ret = False
    current_version = await hub.idemenv.ops.get_current_version()
    override_version_file = Path(os.getcwd()) / ".idem-version"

    if current_version[0]:
        async with aiofiles.open(override_version_file, "w") as ofile:
            await ofile.write(current_version[0])
        ret = True

    return ret


async def remove_version(hub, version, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    idem_bin = hub.idemenv.ops.LOCAL_VERSIONS.get(version)
    if idem_bin:
        idem_bin.unlink()

    return True


async def use_version(hub, version, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    if not hub.idemenv.ops.LOCAL_VERSIONS:
        await hub.idemenv.ops.fill_local_version_list()

    if version in hub.idemenv.ops.LOCAL_VERSIONS:
        version_file = Path(hub.OPT.idemenv.idemenv_dir) / "version"
        version_file.parent.mkdir(parents=True, exist_ok=True)

        async with aiofiles.open(version_file, "w") as vfile:
            await vfile.write(version)
    else:
        print(f"{version} is not installed. Run 'idemenv install {version}' first.")

    return True



def __init__(hub):
    # Remember not to start your app in the __init__ function
    # This function should just be used to set up the plugin subsystem
    # The run.py is where your app should usually start
    for dyne in ["acct", "exec", "tool"]:
        hub.pop.sub.add(dyne_name=dyne)
    for dyne in ["exec", "tool"]:
        hub.pop.sub.load_subdirs(getattr(hub, dyne), recurse=True)


def cli(hub):
    hub.pop.config.load(["idemenv"], cli="idemenv")
    # Your app's options can now be found under hub.OPT.idemenv
    kwargs = dict(hub.OPT.idemenv)

    # Initialize the asyncio event loop
    hub.pop.loop.create()

    # Start the async code
    coroutine = hub.idemenv.init.run(**kwargs)
    hub.pop.Loop.run_until_complete(coroutine)


async def run(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    if hub.SUBPARSER == "init":
        return await hub.idemenv.cli.init()
    elif hub.SUBPARSER == "install":
        return await hub.idemenv.cli.install()
    elif hub.SUBPARSER == "list":
        return await hub.idemenv.cli.list()
    elif hub.SUBPARSER == "list-remote":
        return await hub.idemenv.cli.list_remote()
    elif hub.SUBPARSER == "pin":
        return await hub.idemenv.cli.pin()
    elif hub.SUBPARSER == "uninstall":
        return await hub.idemenv.cli.uninstall()
    elif hub.SUBPARSER == "use":
        return await hub.idemenv.cli.use()
    elif hub.SUBPARSER == "version":
        return await hub.idemenv.cli.version()
    else:
        print(hub.args.parser.help())
        return 2



import os

# https://pop.readthedocs.io/en/latest/tutorial/quickstart.html#adding-configuration-data

# In this dictionary goes all the immutable values you want to show up under hub.OPT.idemenv
CONFIG = {
    "config": {
        "default": None,
        "help": "Load extra options from a configuration file onto hub.OPT.idemenv",
    },
    "repo_url": {
        "default": "https://repo.idemproject.io/idem/",
        "help": "Idem single binary repository location. Version directories are expected here.",
    },
    "idemenv_dir": {
        "default": os.path.join(os.environ["HOME"], ".idemenv"),
        "help": "Working directory for idemenv downloads",
    },
}

# The selected subcommand for your cli tool will show up under hub.SUBPARSER
# The value for a subcommand is a dictionary that will be passed as kwargs to argparse.ArgumentParser.add_subparsers
SUBCOMMANDS = {
    "init": {},
    "install": {},
    "list": {},
    "list-remote": {},
    "pin": {},
    "uninstall": {},
    "use": {},
    "version": {},
}

# Include keys from the CONFIG dictionary that you want to expose on the cli
# The values for these keys are a dictionaries that will be passed as kwargs to argparse.ArgumentParser.add_option
CLI_CONFIG = {
    "config": {"options": ["-c"]},
    # "my_option1": {"subcommands": ["A list of subcommands that exclusively extend this option"]},
    # This option will be available under all subcommands and the root command
    "repo_url": {"options": ["-r", "--repo-url"], "subcommands": ["_global_"]},
    "idemenv_dir": {"options": ["-d", "--idemenv-dir"], "subcommands": ["_global_"]},
    "idem_version": {
        "display_priority": 0,
        "positional": True,
        "subcommands": ["install", "use", "uninstall"],
        "help": "The version of Idem to act on.",
    },
}

# These are the namespaces that your project extends
# The hub will extend these keys with the modules listed in the values
DYNE = {"idemenv": ["idemenv"]}
